/*处理非活动连接*/
#include"head.hpp"
#include"lst_timer.hpp"
#include<signal.h>
#define FD_LIMIT 65535
#define MAX_EVENT_NUMBER 1024
#define TIMESLOT 5

/*统一事件源处理信号 创建管道加入到epoll中*/
static int pipefd[2];
static sort_timer_lst timer_lst;
static int epollfd = 0;

/*将socket设置为非阻塞的*/
int setnonblocking(int fd)
{
    int old_option = fcntl(fd, F_GETFL);
    int new_option = old_option | O_NONBLOCK;
    fcntl(fd, F_SETFL, new_option);
    return old_option;
}

/*设置epoll参数*/
void addfd(int epollfd, int fd)
{
    epoll_event event;
    event.data.fd = fd;
    event.events = EPOLLIN | EPOLLET;
    epoll_ctl(epollfd, EPOLL_CTL_ADD, fd, &event);
    setnonblocking(fd);
}

void sig_handler(int sig)
{
    int save_errno = errno;
    int msg = sig;
    send(pipefd[1], (char *)&msg, 1, 0);
    errno = save_errno;   
}

void addsig(int sig)
{
    struct sigaction sa;
    memset(&sa, '\0', sizeof(sa));
    sa.sa_handler = sig_handler;
    sa.sa_flags |= SA_RESTART;
    sigfillset(&sa.sa_mask);
    assert(sigaction(sig, &sa, NULL) != -1);
}

void timer_handler()
{
    /*定时处理任务，实际上就是调用tick函数*/
    timer_lst.tick();
    /*因为一次alarm调用只会引起一次SIGALRM信号，所以我们重新定时，以不断触发SIGALRM信号*/
    alarm(TIMESLOT);
}

/*定时器回调函数，它删除非活动连接socket上的注册事件，并关闭之*/
void cb_func(client_data* user_data)
{
    epoll_ctl(epollfd, EPOLL_CTL_DEL, user_data->sockfd, 0);
    assert(user_data);
    close(user_data->sockfd);
    printf("close fd %d\n", user_data->sockfd);
}

int main(int argc, char* argv[])
{
    if(argc <= 2)
        my_err("command list", __LINE__);
    const char* ip = argv[1];
    int port = atoi(argv[2]);

    int ret = 0;
    struct sockaddr_in address;
    memset(&address, 0, sizeof(address));
    address.sin_family = AF_INET;
    address.sin_port = htonl(port);
    inet_pton(AF_INET, ip, &address.sin_addr);

    int listenfd = socket(AF_INET, SOCK_STREAM, 0);
    if(listenfd < 0)
        my_err("socket", __LINE__);
    ret = bind(listenfd, (struct sockaddr*)&address, sizeof(address));
    if(ret < 0)
        my_err("bind", __LINE__);
    ret = listen(listenfd, 5);
    if(ret < 0)
        my_err("listen", __LINE__);
    
    epoll_event events[MAX_EVENT_NUMBER];
    epollfd = epoll_create(5);

    addfd(epollfd, listenfd);
    
    /*设置信号处理函数*/
    addsig(SIGALRM);
    addsig(SIGTERM);

    bool stop_server = false;

    client_data* users = new client_data[FD_LIMIT];
    bool timeout = false;
    alarm(TIMESLOT); //定时

    while(!stop_server)
    {
        int number = epoll_wait(epollfd, events, MAX_EVENT_NUMBER, -1); //阻塞
        if( (number < 0) && (errno != EINTR) )
            my_err("epoll_wait", __LINE__);
        for(int i = 0; i < number; ++i)
        {
            int sockfd = events[i].data.fd;
            /*处理新到客户连接 */
            if(sockfd == listenfd)
            {
                struct sockaddr_in client;
                memset(&client, 0, sizeof(client));
                socklen_t client_len = sizeof(client);
                int connfd = accept(listenfd, (struct sockaddr*)&client, &client_len);
                //将新的连接加入到epoll事件集中
                addfd(epollfd, connfd);
                users[connfd].address = client;
                users[connfd].sockfd = connfd;
                /*创建定时器，设置其回调函数与超时时间，然后绑定定时器与用户数据，最后将定时器添加到链表lst中*/
                util_timer* timer = new util_timer;
                timer->user_data = &users[connfd];
                timer->cb_func = cb_func;
                time_t cur = time(NULL);
                timer->expier = cur +  3 * TIMESLOT;
                users[connfd].timer = timer;
                timer_lst.add_timer(timer);
                
            }

            /*处理信号*/
            else if( (sockfd == pipefd[0]) && (sockfd & EPOLLIN))
            {
                int sig;
                char signals[1024];
                ret = recv(sockfd, signals, sizeof(signals), 0);
                if(ret == -1)
                {
                    continue;
                }      
                else if(ret == 0)
                    continue;
                else
                {
                    for(int i = 0; i < ret; ++i)
                    {
                        switch(signals[i])
                        {
                            case SIGALRM:
                            {
                                /*用timeout变量标记有定时任务需要处理，但不立即处理定时任务。
                                这是因为定时任务的优先级不是很高，我们优先处理其它更重要的任务*/
                                timeout = true;
                                break;
                            }
                            case SIGTERM: 
                            {
                                stop_server = true;
                            }
                        }
                    }
                }
            }
            
            /*处理客户连接上收到的数据*/
            else if(sockfd & EPOLLIN)
            {
                memset(users[sockfd].buf, '\0', BUFFER_SIZE);
                ret = recv(sockfd, users[sockfd].buf, BUFFER_SIZE - 1, 0);
                printf("get %d bytes of client data %s from %d\n", ret, users[sockfd].buf, sockfd);

                util_timer* timer = users[sockfd].timer;
                if(ret < 0)
                {
                    /*如果发生读错误，则关闭连接，并移除其对应的定时器*/
                    if(errno != EAGAIN)
                    {
                        cb_func(&users[sockfd]);
                        if(timer)
                            timer_lst.del_timer(timer);
                    }
                }
                else if(ret == 0)
                {
                    /*如果对方已经关闭连接，则我们也关闭连接，并移除对应的定时器*/
                    cb_func(&users[sockfd]);
                    if(timer)
                        timer_lst.del_timer(timer);
                }
                else
                {
                    /*如果某个客户连接上有数据可读，则我们要调整该连接对应的定时器，以延迟该连接被关闭时间*/
                    if(timer)
                    {
                        time_t cur = time(NULL);
                        timer->expier = cur + 3 * TIMESLOT;
                        printf("adjust time once\n");
                        timer_lst.adjust_timer(timer);
                    }
                }
                
            }

            
        }
        /*最后处理定时事件，因为I/O事件有更高的优先级。当然，这样做将导致定时任务不能精确地按照预期的时间执行*/
        if(timeout)
        {
            timer_handler();
            timeout = false;
        }
    }

    close(listenfd);
    close(pipefd[0]);
    close(pipefd[1]);
    delete []users;
    return 0;
}


